from pionex_py.internal.RestClient import RestClient
from pionex_py.internal.PionexExceptions import assert_valid_market_type

class Markets(RestClient):
    def __init__(self):
        super().__init__()

    def get_trades(self, symbol: str, limit: int = None):
        return self._send_request('GET', '/api/v1/market/trades', symbol=symbol, limit=limit)

    def get_depth(self, symbol: str, limit: int = None):
        return self._send_request('GET', '/api/v1/market/depth', symbol=symbol, limit=limit)

    def get_24hr_ticker(self, symbol: str = None, type: str = None):
        assert_valid_market_type(type)
        return self._send_request('GET', '/api/v1/market/tickers', symbol=symbol, type=type)

    def get_book_ticker(self, symbol: str = None, type: str = None):
        assert_valid_market_type(type)
        return self._send_request('GET', '/api/v1/market/bookTickers', symbol=symbol, type=type)

    def get_klines(self, symbol: str, interval: str, endTime: int = None, limit: int = None):
        return self._send_request('GET', '/api/v1/market/klines', symbol=symbol, interval=interval, endTime=endTime, limit=limit)



class PionexException(Exception):
    pass

class WebsocketError(PionexException):
    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return self.error_message

class REST_Exception(PionexException):
    def __init__(self, response):
        self.code = response['code']
        self.message = response['message']

    def __str__(self):
        return f"{self.code} \n[!] {self.message}"


def assert_valid_market_type(type_str):
    if type_str:
        assert type_str in ['SPOT','PERP'], "unknown type"



from pionex_py.internal.RestClient import RestClient

class Orders(RestClient):
    def __init__(self, key, secret):
        super().__init__(key, secret)

    @staticmethod
    def assert_valid_order_side_type(side, type):
        assert side in ['BUY', 'SELL'], "unknown side"
        assert type in ['LIMIT', 'MARKET'], "unknown type"

    @staticmethod
    def assert_valid_mass_orders(orders):
        for order in orders:
            assert order['side'] in ['BUY', 'SELL'], "unknown side"
            assert order['type'] == 'LIMIT', "mass orders only support LIMIT"
            assert order['size'], "order needs to have a size"
            assert order['price'], "limit oder needs to have a price"

    def new_order(self, symbol: str, side: str, type: str, clientOrderId: str = None, size: str = None, price: str = None, amount: str = None, IOC: bool = None):
        self.assert_valid_order_side_type(side, type)
        data = {
          "symbol": symbol,
          "side": side,
          "type": type,
          "clientOrderId": clientOrderId,
          "size": size,
          "price": price,
          "amount": amount,
          "IOC": IOC,
        }
        return self._send_request('POST', '/api/v1/trade/order', data=data)

    def new_multiple_order(self, symbol: str, orders: list = None):
        self.assert_valid_mass_orders(orders)
        data = {
            "symbol": symbol,
            "orders": orders
        }
        return self._send_request('POST', '/api/v1/trade/massOrder', data=data)

    def get_order(self, orderId: int):
        return self._send_request('GET', '/api/v1/trade/order', orderId=orderId)

    def get_order_by_client_order_id(self, clientOrderId: int):
        return self._send_request('GET', '/api/v1/trade/orderByClientOrderId', clientOrderId=clientOrderId)

    def cancel_order(self, symbol: str, orderId: int):
        data = {'symbol':symbol, 'orderId':orderId}
        return self._send_request('DELETE', '/api/v1/trade/order', data=data)

    def get_open_orders(self, symbol: str):
        return self._send_request('GET', '/api/v1/trade/openOrders', symbol=symbol)

    def get_all_orders(self, symbol: str, startTime: int = None, endTime: int = None, limit: int = None):
        return self._send_request('GET', '/api/v1/trade/allOrders', symbol=symbol, startTime=startTime, endTime=endTime, limit=limit)

    def get_fills(self, symbol: str, startTime: int = None, endTime: int = None):
        return self._send_request('GET', '/api/v1/trade/fills', symbol=symbol, startTime=startTime, endTime=endTime)

    def get_fills_by_order_id(self, orderId: int, fromId: int = None):
        return self._send_request('GET', '/api/v1/trade/fillsByOrderId', orderId=orderId, fromId=fromId)

    def cancel_all_orders(self, symbol: str):
        data = {'symbol':symbol}
        return self._send_request('DELETE', '/api/v1/trade/allOrders', data=data)



from pionex_py.internal.RestClient import RestClient

class Account(RestClient):
    def __init__(self, key, secret):
        super().__init__(key, secret)

    def get_balance(self, coin=None):
        if coin is None:
            return self._send_request('GET', '/api/v1/account/balances')
        else:
            result = self._send_request('GET', '/api/v1/account/balances')
            return next(
                (balance['free'] for balance in result['data']['balances'] if balance['coin'] == coin), 
                0
            )
        


from pionex_py.internal.generate_signature import rest_signature
from pionex_py.internal.PionexExceptions import REST_Exception
from pionex_py import __version__

import requests
import time
import json

class RestClient:
    def __init__(self, key:str='', secret:str=''):
        self.base_url = 'https://api.pionex.com/'
        self.key = key
        self.secret = secret
        self.session = requests.Session()
        self.session.headers.update({
            'User-Agent': f'pionex-python/{__version__}',
            'Content-Type': 'application/json;charset=utf-8',
        })
        if key:
            self.session.headers.update({
                'PIONEX-KEY':key
            })

    def _send_request(self, http_method, url_path, **params):
        url = self.base_url + url_path
        params = {k: v for k, v in params.items() if v is not None}  # Remove None params

        timestamp = str(int(time.time() * 1000))
        params['timestamp'] = timestamp

        if 'data' in params:
            data = params['data']
            del params['data']
        else:
            data = None

        if self.key:
            signature = rest_signature(self.secret, http_method, url_path, timestamp, params, data)
            self.session.headers.update({
                'PIONEX-KEY': self.key,
                'PIONEX-SIGNATURE': signature
            })

        match(http_method):
            case 'GET':
                r = self.session.get(url, params=params, json=data)
            case 'DELETE':
                r = self.session.delete(url, params=params, json=data)
            case 'PUT':
                r = self.session.put(url, params=params, json=data)
            case 'POST':
                r = self.session.post(url, params=params, json=data)
            case _:
                raise REST_Exception('Unsupported HTTP method')

        response = r.json()

        if not response['result']:
            raise REST_Exception(response)
        return response



from pionex_py.internal.generate_signature import ws_signature
from pionex_py.internal.WebsocketClient import WebsocketClient

from pionex_py.websocket.PublicStream import PublicStream

import time

class PrivateStream(PublicStream):
    """
    Extends PublicStream with authentication
    """
    def __init__(self, api_key:str, api_secret:str):
        self.key_secret = (api_key, api_secret)
        super().__init__()

    def initialize_client(self):
        key, secret = self.key_secret
        timestamp = str(int(time.time() * 1000))
        signature = ws_signature(key, secret, timestamp)
        url = f"wss://ws.pionex_python.com/ws?key={key}&timestamp={timestamp}&signature={signature}"
        return WebsocketClient(url, self._on_message)



from pionex_py.internal.RestClient import RestClient
from pionex_py.internal.PionexExceptions import assert_valid_market_type

#rename python function so we can have "type" argument
type_func = type

class Common(RestClient):
    def __init__(self):
        super().__init__()

    def market_data(self, symbols:list[str]=None, market_type='PERP'):
        """
        Fetches market data for all symbols or a specific symbol.
        
        Parameters:
            symbols (list[str], optional): List of symbols. (Optional)
            type (str, optional): PERP, SPOT. (Optional)
        Returns:
            Response object from the API request.
        """
        assert_valid_market_type(market_type)
        if type_func(symbols) == list:
            symbols = ','.join(symbols)
        return self._send_request('GET', '/api/v1/common/symbols', symbols=symbols, type=market_type)

    def klines(self, symbol:str, interval:str, endTime:int=None, limit:int=None):
        """
        Fetches kline (candlestick) data for a given symbol and interval.
        
        Parameters:
            symbol (str): Symbol. (Mandatory)
            interval (str): 1M, 5M, 15M, 30M, 60M, 4H, 8H, 12H, 1D. (Mandatory)
            endTime (int, optional): End time in milliseconds. (Optional)
            limit (int, optional): Default 100, range: 1-500. (Optional)
        Returns:
            Response object from the API request.
            
        Example: https://api.pionex.com/api/v1/market/klines?symbol=BTC_USDT&interval=5M
        """
        return self._send_request('GET', '/api/v1/market/klines', symbol=symbol, interval=interval, endTime=endTime, limit=limit)


import json
import hashlib
import hmac
from urllib.parse import urlencode

# generate the PIONEX-SIGNATURE

# https://pionex-doc.gitbook.io/apidocs/restful/general/authentication
def rest_signature(api_secret, method, path, timestamp, params=None, data=None):
    params = params if params else {'timestamp':timestamp}

    query_string = urlencode(sorted(params.items()))
    
    message = f"{method.upper()}{path}?{query_string}"

    if data:
        message += json.dumps(data, separators=(', ', ': '))

    signature = hmac.new(api_secret.encode('utf-8'), message.encode('utf-8'), hashlib.sha256).hexdigest()

    return signature

# https://pionex-doc.gitbook.io/apidocs/websocket/general-info
def ws_signature(api_key, api_secret, timestamp):
    params = {
        'key':api_key,
        'timestamp':timestamp
    }
    query_string = urlencode(sorted(params.items()))
    message = f"/ws?{query_string}websocket_auth"

    signature = hmac.new(api_secret.encode('utf-8'), message.encode('utf-8'), hashlib.sha256).hexdigest()

    return signature


from pionex_py.internal.WebsocketClient import WebsocketClient
from pionex_py.internal.PionexExceptions import WebsocketError

from typing import Callable
import time

class PublicStream:
    """
    subscribe to multiple pionex public websocket streams
    and assign individual callbacks for each unique (topic, symbol) pair
    """
    def __init__(self):
        self.ws_client = self.initialize_client()
        self.ws_client.start()
        self.topic_symbol_callbacks = {}

    def initialize_client(self):
        return WebsocketClient("wss://ws.pionex.com/wsPub", self._on_message)

    def _on_message(self, message:dict):
        if 'code' in message:
            raise WebsocketError(message['message'])
        topic = None
        if 'topic' in message:
            topic = message['topic']
        symbol = None
        if 'symbol' in message:
            symbol = message['symbol']
        if (topic, symbol) in self.topic_symbol_callbacks:
            if 'data' in message:
                self.topic_symbol_callbacks[(topic, symbol)](message['data'])

    def subscribe(self, callback:Callable[[dict], any], topic:str, symbol:str = None, limit:int = None):
        message = {
            'op': "SUBSCRIBE",
            'topic': topic
        }
        if symbol != None:
            message['symbol'] = symbol
        if limit != None:
            message['limit'] = limit
        self.ws_client.send_message(message)
        self.topic_symbol_callbacks[(topic, symbol)] = callback

    def unsubscribe(self, topic:str, symbol:str=None):
        message = {
            'op': "UNSUBSCRIBE",
            'topic': topic
        }
        if symbol != None:
            message['symbol'] = symbol
        self.ws_client.send_message(message)

    def __del__(self):
        self.ws_client.close()
        del self.ws_client



import threading
import logging
import json
import time
from typing import Callable
from websocket import (
    ABNF,
    create_connection,
    WebSocketException,
    WebSocketConnectionClosedException,
    WebSocketTimeoutException,
)

class WebsocketClient(threading.Thread):
    """
    implements thread to run websocket send and recv
    - handles opening and closing
    - handles heartbeat ping/pong
    - callback on_message()
    - send individual message
    """
    def __init__(self, url, on_message:Callable[[dict], any]):
        threading.Thread.__init__(self)
        self.logger = logging.getLogger(__name__)
        self.on_message = on_message
        self.create_ws_connection(url)

    def create_ws_connection(self, url:str):
        self.ws = create_connection(url)
        self.logger.debug(f"WebSocket connection has been established to: {url}",)

    def run(self):
        self._read_data()

    def send_message(self, message: dict):
        message = json.dumps(message, separators=(', ', ': '))
        self.logger.debug(f"Sending message to Pionex: {message}")
        self.ws.send(message)

    def _read_data(self):
        while True:
            try:
                op_code, frame = self.ws.recv_data_frame(True)
                self.logger.debug(f"received message: {frame.data.decode('utf-8')}")
            except WebSocketException as e:
                if isinstance(e, WebSocketConnectionClosedException):
                    self.logger.error("Unexpected websocket close connection")
                elif isinstance(e, WebSocketTimeoutException):
                    self.logger.error("Websocket connection timeout")
                else:
                    self.logger.error(f"Websocket exception: {e}")
                raise e
            except Exception as e:
                self.logger.error(f"Exception in read_data: {e}")
                raise e

            if op_code == ABNF.OPCODE_PING:
                self.ws.pong("")

            if op_code == ABNF.OPCODE_BINARY:
                data = json.loads(frame.data)
                if 'op' in data:
                    if data['op'] == 'PING':
                        pass
                        self._pong()
                    elif data['op'] == 'CLOSE':
                        if 'note' in data:
                            if data['note'] == 'closed by yourself':
                                break
                        self.logger.warning("forced op:CLOSE by server")
                        break
                else:
                    self.on_message(data)

    def _pong(self):
        ms_timestamp = int(time.time() * 1000)
        message = {'op':'PONG','timestamp':ms_timestamp}
        self.send_message(message)

    def close(self):
        if not self.ws.connected:
            self.logger.warning("Websocket already closed")
        else:
            self.ws.send_close()
            self.logger.debug("Closing connection")
        return